#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <mpi.h>

#define DIM		201
#define H		0.01       	 /* grid spacing */
#define EPS		0.001

/* communication tags */
#define RIGHT		101
#define LEFT		202

double heat (double x, double y) {     /* exact values for boundary */
  return -4.7 + 0.2*x - 0.3*y + 0.09*(x*x + y*y) +
    log(sqrt((x + 0.2)*(x + 0.2) + (y + 0.2)*(y + 0.2)));
}

void main (int argc, char *argv[]) {
  double w = 1.95;		/* SOR parameter */
  double maximum, maxdelta, tmp, grid[DIM][DIM];
  double buf[DIM-2];		/* buffer for send and receive */
  int id, procs, i, j, c, iter = 0, t0, t1, x1, x2;
  MPI_Status status;

  MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &id);
  MPI_Comm_size(MPI_COMM_WORLD, &procs);

/* run the program for 10 different values for w */
  for(w = 1.9; w < 2.0; w += 0.01) {
    iter = 0;

    for (i=0; i<DIM; i++) {
      grid[0][i]	= heat(0.0, H*i);
      grid[DIM-1][i]	= heat(H*(DIM-1), H*i);
      grid[i][0]	= heat(H*i, 0.0);
      grid[i][DIM-1]	= heat(H*i, H*(DIM-1));
    }

    for (i=1; i<DIM-1; i++)
      for (j=1; j<DIM-1; j++)
	grid[i][j] = 0.0;

/* Determine the leftmost and rightmost x coordinate that this processor
    must compute. The formulas are correct if procs divides DIM. */
    x1 = id * DIM / procs;
    x2 = (id + 1) * DIM / procs - 1;

    if (id == 0) x1++;
    if (id == procs-1) x2--;

/* printf("Processor %d out of %d from %d to %d\n", id, procs, x1, x2); */

    t0 = time(0);

    for (;;) {
      maxdelta = 0.0;
      for (c=0; c<=1; c++) {		/* red & black phases */
	for (i=x1; i<=x2; i++)
	  for (j=2-c+(2*c-1)*(i&1); j <= DIM-2; j+=2) {
	    tmp = grid[i][j];
	    grid[i][j] = 0.25 * w * (grid[i-1][j] + grid[i+1][j] +
			     grid[i][j-1] + grid[i][j+1]) + (1.0 - w) * tmp;
	    if (fabs(grid[i][j] - tmp) > maxdelta)
	      maxdelta = fabs(grid[i][j] - tmp);
				}

/* now send all the updated values to the respective neighbor processors,
	first to the right, then to the left. Must be done after each phase. */

	if (id != procs-1) {
	  for(j = 1; j <= DIM-2; j++) buf[j-1] = grid[x2][j];
	  MPI_Send(&buf[0], DIM-2, MPI_DOUBLE, id+1, RIGHT, MPI_COMM_WORLD);

	  MPI_Recv(&buf[0], DIM-2, MPI_DOUBLE, id+1, LEFT, MPI_COMM_WORLD, &status);
	  for(j = 1; j <= DIM-2; j++) grid[x2+1][j] = buf[j-1];
	}

	if (id != 0) {
	  MPI_Recv(&buf[0], DIM-2, MPI_DOUBLE, id-1, RIGHT, MPI_COMM_WORLD, &status);
	  for(j = 1; j <= DIM-2; j++) grid[x1-1][j] = buf[j-1];
	  
	  for(j = 1; j <= DIM-2; j++) buf[j-1] = grid[x1][j];
	  MPI_Send(&buf[0], DIM-2, MPI_DOUBLE, id-1, LEFT, MPI_COMM_WORLD);
	}

      }

      iter++;

/* maxdelta holds the maximum difference between successive values
    on each processor. Now determine the global maximum. */
      MPI_Allreduce(&maxdelta, &maximum, 1, MPI_DOUBLE, MPI_MAX, MPI_COMM_WORLD);

/* maximum holds the GLOBAL maximum, so all processors will
    terminate at the same time. */
		if (maximum < EPS) break;

/* if (id == 0) printf("maxdelta : %f\n", maximum); */
    }

    t1 = time(0);

    if (id == 0) {
      printf("%d iterations in %d seconds on %d processors\n", iter, t1-t0, procs);
      printf("dim %d    w %f    epsilon %f\n", DIM, w, maxdelta);
    }

  }
  MPI_Finalize();
}
