/*  ------------------------------------------------------------------------
 *      wave_mw.c
 *      Lab exercise for More on Collective Communication
 *
 *      This program implements the concurrent wave equation described  
 *      in Chapter 5 of Fox et al., 1988, Solving Problems on Concurrent
 *      Processors, vol 1.   
 *
 *      A vibrating string is decomposed into points.  Each task is  
 *      responsible for updating the amplitude of a number of points over
 *      time.
 *       
 *      At each iteration, each task exchanges boundary points with
 *      nearest neighbors.  This version uses low level sends and receives
 *      to exchange boundary points.
 *
 *      AUTHOR: R. Leibensperger 6/93 
 *      COVERTED TO MPI: Xianneng Shen 11/94
 *      ADAPTED FOR MORE ON COLLECTIVE COMMUNICATION: R. Leibensperger 4/96
 *      IMPROVEMENT TO ENDPOINT EXCHANGE: R. Leibensperger 8/30
 *  ------------------------------------------------------------------------ */

#include <stdlib.h>
#include <math.h>
#include <stdio.h>
#include "mpi.h"
#define MASTER 0 
#define MAXPROC 8
#define MAXPOINTS 1000
#define MAXSTEPS 10000
#define PI 3.14159265

void get_data(void);
void distribute_data(void);
void update(int left, int right);
void collect_data(void);
void out_master(void);
MPI_Status status;
MPI_Request request;
int rank,                    /* rank in MPI_COMM_WORLD */
    ntask,                   /* number of tasks */
    nsteps,                  /* number of time steps */
    tpoints,                 /* total points along string */
    npoints[MAXPROC],        /* number of points handled by each task */
    displ[MAXPROC],          /* displacement of 1st point handled by each */
                             /* task from start of allval array */
    rcode;                   /* generic return code */
double allval[MAXPOINTS],    /* complete array of amplitudes */
       values[MAXPOINTS+2],  /* values at time t, block distributed to */
                             /* each task */
       oldval[MAXPOINTS+2],  /* values at time (t-dt) */
       newval[MAXPOINTS+2];  /* values at time (t+dt) */

/*  ------------------------------------------------------------------------
 *      Master obtains input values from user
 *  ------------------------------------------------------------------------ */

void get_data(void)
{
  int buffer[2];
  FILE *fdesc;

  /* Master reads in number of points, number of time steps, */
  /* and initial data values */
  if (rank == MASTER) 
    {
      fdesc = fopen("c.string.data","r");
      fread (&tpoints, sizeof(int), 1, fdesc);
      fread (&nsteps, sizeof(int), 1, fdesc);
      fread (allval, sizeof(double), tpoints, fdesc);
      fclose (fdesc);
      printf("%d: points = %d, steps = %d\n", rank, tpoints, nsteps);
      buffer[0] = tpoints;
      buffer[1] = nsteps;
    }

  /* Broadcast number of points and time steps */
  MPI_Bcast(&buffer, 2, MPI_INT, MASTER, MPI_COMM_WORLD);
  tpoints = buffer[0];
  nsteps = buffer[1];
}

/*  ------------------------------------------------------------------------
 *     All tasks participate in a scatterv to distribute data
 *  --------------------------------------------------------------------- */

void distribute_data (void)
{
  int nmin, nleft, i, k;

  /*  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  * 
   *   At this point in the program, all tasks know:                    *
   *       rank = their rank in MPI_COMM_WORLD                          *
   *       tpoints = the total number of points on the wave             *
   *       ntask = the number of tasks participating in the computation *
   *   In addition, the master knows:                                   *
   *       allval[MAXPOINTS] = initial data (array of doubles)          *
   *                                                                    *
   *   First fill the following arrays:                                 *
   *      npoints[ntask] = number of points assigned to each task       *
   *      displ[ntask] = displacement in allval of first point          *
   *                     assigned to each task                          *
   *                                                                    *
   *   Then scatter the array allval to the tasks. Each task should     *
   *   store its block in values[1] to values[npoints].  Values[0] and  *
   *   values[npoints+1] will be needed for storing boundary values.    *
   *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  */

  /* Determine number of points assigned to each task and displacement */
  /* of first point assigned to each task */

  printf ("%d: displ = %d, npoints = %d\n", rank, displ[rank], 
	  npoints[rank]);

  /* call MPI_Scatterv */

}

/*  -------------------------------------------------------------------------
 *      Calculate new values using wave equation
 *  -------------------------------------------------------------------------*/
  
void do_math(int i)
{
  const double dtime = 0.3;
  const double c = 1.0;
  const double dx = 1.0; 
  double tau, sqtau;

  tau = (c * dtime / dx);
  sqtau = tau * tau;
  newval[i] = (2.0 * values[i]) - oldval[i]  
    + (sqtau * (values[i-1] - (2.0 * values[i]) + values[i+1]));
}

/*  -------------------------------------------------------------------------
 *      All tasks update their points a specified number of times  
 *  -------------------------------------------------------------------------*/
  
void update(int left, int right)
{
  int i, j;
  int E_RtoL = 10;
  int E_LtoR = 20;
  
  for (i = 1; i <= npoints[rank]; i++) 
    oldval[i] = values[i];

  /* update values for each point along string */
  for (i = 1; i <= nsteps; i++)
    {
      /* Send to the left, receive from the right */
      /* End nodes have "null" neighbors -- this prevents communication */
      /* from wrapping */
      MPI_Isend(&values[1], 1, MPI_DOUBLE, left, E_RtoL,
		MPI_COMM_WORLD, &request);
      MPI_Recv(&values[npoints[rank]+1], 1, MPI_DOUBLE, right,
	       E_RtoL, MPI_COMM_WORLD, &status);
      MPI_Wait(&request, &status);

      /*  Send to the right, receive from the left */
      MPI_Isend(&values[npoints[rank]], 1, MPI_DOUBLE, right,
		E_LtoR, MPI_COMM_WORLD, &request);
      MPI_Recv(&values[0], 1, MPI_DOUBLE, left, E_LtoR,
               MPI_COMM_WORLD, &status);
      MPI_Wait(&request, &status);

      /* update points along line */
      for (j = 1; j <= npoints[rank]; j++)
	{
	  /* global endpoints */
	  if ((displ[rank] + j - 1 == 0) || 
	      (displ[rank] + j == tpoints))
	    newval[j] = 0.0;
	  else
	    do_math(j);
	}
      for (j = 1; j <= npoints[rank]; j++)
	{
	  oldval[j] = values[j];
	  values[j] = newval[j];
	}
    }
}

/*  ------------------------------------------------------------------------
 *      All tasks participate in a gatherv to collect data
 *  ------------------------------------------------------------------------ */
void collect_data(void)
{
  /*  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *
   *    Each task has now updated its data in the values array.            *
   *    Gather this data into the allval array on the task with rank 0.    *
   *    This data should be stored from allval[0] to allval[tpoints].      *
   *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  *  */

}

/*  ------------------------------------------------------------------------
 *      Master receives results from workers and prints
 *  ------------------------------------------------------------------------ */

void out_master(void)  
{
  int istep, i;
  FILE *fout;
  
  /* print 10 points for validation */
  istep = (tpoints <= 10) ? 1: tpoints/10;
  printf ("Points for validation:\n");
  for (i = 0; i < tpoints; i+=istep)
    printf ("%d:%4.2f  ", i, allval[i]);
  if (i-istep != tpoints - 1) 
    printf ("%d:%4.2f  ", tpoints-1, allval[tpoints-1]);
  printf("\n");
  
  fout = fopen ("c.output.data", "w");
  fwrite (allval, sizeof(double), tpoints, fout);
  fclose (fout);
}

/*  ------------------------------------------------------------------------
 *      Main program
 *  ------------------------------------------------------------------------ */

main(int argc, char **argv)
{
  int left, right;

  /* learn number of tasks and rank in MPI_COMM_WORLD */
  rcode = MPI_Init(&argc, &argv);
  MPI_Comm_rank(MPI_COMM_WORLD, &rank);
  MPI_Comm_size(MPI_COMM_WORLD, &ntask);
  if (rcode >= 0) printf ("%d: Wave Program running\n", rank);

  /* determine left and right neighbors */
  /* Communication will not wrap, so end nodes have "null" neighbors */
  if (rank == ntask-1) right = MPI_PROC_NULL;
  else right = rank + 1;
  if (rank == 0) left = MPI_PROC_NULL;
  else left = rank - 1;

  /* get program parameters and initialize wave values */
  get_data();
  distribute_data();

  /* update values along the line for nstep time steps */
  update(left, right);

  /* Results are collected, and master write to file */
  collect_data();
  if (rank == MASTER) out_master();

  MPI_Finalize();
}
