/*
 * File         : Datatype.java
 * Author       : Sang Lim, Sung-Hoon Ko, Xinying Li
 * Created      : Thu Apr  9 12:22:15 1998
 * Revision     : $Revision: 1.4 $
 * Updated      : $Date: 1999/03/03 19:55:37 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

package mpi;
//import mpi.*;

public class Datatype {
  private final static int UNDEFINED = -1;
  private final static int NULL    = 0;
  private final static int BYTE    = 1;
  private final static int CHAR    = 2;
  private final static int SHORT   = 3;
  private final static int BOOLEAN = 4;
  private final static int INT     = 5;
  private final static int LONG    = 6;
  private final static int FLOAT   = 7;
  private final static int DOUBLE  = 8;  
  private final static int PACKED  = 9;
  private final static int LB      =10;
  private final static int UB      =11;
  private final static int OBJECT  = 12;

  private static native void init();

  public Datatype(int Type) { 
    GetDatatype(Type); 
    baseType = Type;
  }

  private Datatype(long _handle, Datatype base) {
    handle = _handle;
    baseType = base.baseType;
  }

  private native void GetDatatype(int Type);


  public boolean isObject(Datatype type) {
    if(type.baseType == OBJECT)
      return true;
    else return false;
  }
  /**
   * Returns the extent of a datatype - the difference between
   * upper and lower bound.
   * <p>
   * <table>
   * <tr><td><em> returns: </em></td><td> datatype extent </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_EXTENT</tt>.
   */

  public native int extent();
  public int Extent(){
    if (baseType == OBJECT)
      return _extent;
    else
      return extent();//this.Size(); 
  }

  /**
   * Returns the total size of a datatype - the number of buffer
   * elements it represents.
   * <p>
   * <table>
   * <tr><td><em> returns: </em></td><td> datatype size </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_SIZE</tt>.
   */

  private native int size();
  public int Size(){
    if (baseType == OBJECT)
      return displacements.length;
    else 
      return this.baseSize();  
  }

  /**
   * Find the lower bound of a datatype - the least value
   * in its displacment sequence.
   * <p>
   * <table>
   * <tr><td><em> returns: </em></td><td> displacement of lower bound
   *                                      from origin </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_LB</tt>.
   */

  private native int lB();
  public int Lb(){
    if (baseType == OBJECT)
      return lb;
    else 
      return lB();
  }

  /**
   * Find the upper bound of a datatype - the greatest value
   * in its displacment sequence.
   * <p>
   * <table>
   * <tr><td><em> returns: </em></td><td> displacement of upper bound
   *                                      from origin </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_UB</tt>.
   */

  private native int uB();
  public int Ub(){
    if (baseType == OBJECT)
      return ub;
    else 
      return uB(); 
  }

  /**
   * Commit a derived datatype.
   * Java binding of the MPI operation <tt>MPI_TYPE_COMMIT</tt>.
   */

  private native void commit();
  public void Commit(){
    if (baseType != OBJECT)
      commit();
  }

  public native void finalize();

  public int baseSize() { return this.size();}

  /**
   * Construct new datatype representing replication of old datatype into
   * contiguous locations.
   * <p>
   * <table>
   * <tr><td><tt> count    </tt></td><td> replication count </tr>
   * <tr><td><tt> oldtype  </tt></td><td> old datatype </tr>
   * <tr><td><em> returns: </em></td><td> new datatype </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_CONTIGUOUS</tt>.
   * <p>
   * The base type of the new datatype is the same as the base type of
   * <tt>oldtype</tt>.
   */

  public static Datatype Contiguous(int      count, 
                                    Datatype oldtype){  

    if (oldtype.baseType == OBJECT){
      Datatype result = new Datatype(0,MPI.OBJECT);
      result.lb = 2147483647;result.ub = -2147483647;
      int temp = 0;
      int ptr = 0, size = 0, i , j, l;

      size = count * oldtype.Size();
        
      result.displacements = new int[size];

      for (i = 0 ;i < count; i++)
        for (l = 0;l < oldtype.Size();l++){
          result.displacements[ptr] = oldtype.displacements[l] + 
                                      oldtype.Extent() * i;

          if (result.displacements[ptr] > result.ub)
            result.ub = result.displacements[ptr];
          if (result.displacements[ptr] < result.lb)
            result.lb = result.displacements[ptr];
          ptr++;
        }
        result._extent = result.ub - result.lb + 1;

        return result; 
    }
    else
      return new Datatype(GetContiguous(count, oldtype),oldtype);      
  }

  private static native long GetContiguous(int count, Datatype oldtype);

  /**
   * Construct new datatype representing replication of old datatype into
   * locations that consist of equally spaced blocks.
   * <p>
   * <table>
   * <tr><td><tt> count       </tt></td><td> number of blocks </tr>
   * <tr><td><tt> blocklength </tt></td><td> number of elements in each
   *                                         block </tr>
   * <tr><td><tt> stride      </tt></td><td> number of elements between
   *                                         start of each block </tr>
   * <tr><td><tt> oldtype     </tt></td><td> old datatype </tr>
   * <tr><td><em> returns:    </em></td><td> new datatype </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_VECTOR</tt>.
   * <p>
   * The base type of the new datatype is the same as the base type of
   * <tt>oldtype</tt>.
   */

  public static Datatype Vector(int      count, 
                                int      blocklength, 
                                int      stride, 
                                Datatype oldtype){
 
    if (oldtype.baseType == OBJECT){
      Datatype result = new Datatype(0,MPI.OBJECT);
      result.lb = 2147483647;result.ub = -2147483647;
      int temp = 0;
      int ptr = 0, size = 0, i , j, l;

      size = count * blocklength * oldtype.Size();

      result.displacements = new int[size];

      for (i = 0 ;i < count; i++)
        for (j = 0; j < blocklength; j++)
          for (l = 0;l < oldtype.Size();l++){
            result.displacements[ptr] = oldtype.displacements[l] +
				        oldtype.Extent() +
                                        (stride * i + j);

            if (result.displacements[ptr] > result.ub)
              result.ub = result.displacements[ptr];
            if (result.displacements[ptr] < result.lb)
              result.lb = result.displacements[ptr];
            ptr++;
          }
      result._extent = result.ub - result.lb + 1;
      return result;   
    }
    else
      return new Datatype(GetVector(count, blocklength, stride,oldtype), 
                          oldtype);
  }

  private static native long GetVector(int      count, 
                                       int      blocklength, 
                                       int      stride,
                                       Datatype oldtype);

  /**
   * Identical to <tt>vector</tt> except that the stride is expressed
   * directly in terms of the buffer index, rather than the units of
   * the old type.
   * <p>
   * <table>
   * <tr><td><tt> count       </tt></td><td> number of blocks </tr>
   * <tr><td><tt> blocklength </tt></td><td> number of elements in each
   *                                         block </tr>
   * <tr><td><tt> stride      </tt></td><td> number of elements between
   *                                         start of each block </tr>
   * <tr><td><tt> oldtype     </tt></td><td> old datatype </tr>
   * <tr><td><em> returns:    </em></td><td> new datatype </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_HVECTOR</tt>.
   * <p>
   * <em>Unlike other language bindings</em>, the value of <tt>stride</tt>
   * is <em>not</em> measured in bytes.
   */

  public static Datatype Hvector(int      count, 
                                 int      blocklength, 
                                 int      stride, 
                                 Datatype oldtype){
 
    if (oldtype.baseType == OBJECT){
      Datatype result = new Datatype(0,MPI.OBJECT);
      result.lb = 2147483647;result.ub = -2147483647;
      int temp = 0;
      int ptr = 0, size = 0, i , j, l;
        
      size = count * blocklength * oldtype.Size();
        
      result.displacements = new int[size];

      for (i = 0 ;i < count ; i++)
        for (j = 0; j < blocklength; j++)
          for (l = 0;l < oldtype.Size();l++){
            result.displacements[ptr] = oldtype.displacements[l] + 
                                        stride * i +
                                        oldtype.Extent() * j;

            if (result.displacements[ptr] > result.ub)
              result.ub = result.displacements[ptr];
            if (result.displacements[ptr] < result.lb)
              result.lb = result.displacements[ptr];
            ptr++;
          }
      result._extent = result.ub - result.lb + 1;

      return result;  
    }
    else
      return new Datatype(GetHvector(count, blocklength, stride,oldtype), 
                          oldtype);
  }

  private static native long GetHvector(int count, int blocklength,
                                        int stride, Datatype oldtype) ;

  /**
   * Construct new datatype representing replication of old datatype into
   * a sequence of blocks where each block can contain a different number
   * of copies and have a different displacement.
   * <p>
   * <table>
   * <tr><td><tt> array_of_blocklengths  </tt></td><td> number of elements per
   *                                                    block </tr>
   * <tr><td><tt> array_of_displacements </tt></td><td> displacement of each
   *                                                    block in units of
   *                                                    old type </tr>
   * <tr><td><tt> oldtype                </tt></td><td> old datatype </tr>
   * <tr><td><em> returns:               </em></td><td> new datatype </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_INDEXED</tt>.
   * <p>
   * The number of blocks is taken to be size of the
   * <tt>array_of_blocklengths</tt> argument.  The second argument,
   * <tt>array_of_displacements</tt>, should be the same size.
   * The base type of the new datatype is the same as the base type of
   * <tt>oldtype</tt>.
   */

  public static Datatype Indexed(int[]    blocklengths, 
                                 int[]    _displacements, 
                                 Datatype oldtype){
  
    if (oldtype.baseType == OBJECT){
      Datatype result = new Datatype(0,MPI.OBJECT);
      result.lb = 2147483647;result.ub = -2147483647;
      int temp = 0;
      int ptr = 0, size = 0, i , j, l;

      for (i = 0; i < blocklengths.length; i++)
        size += blocklengths[i] * oldtype.Size();
  
      result.displacements = new int[size];

      for (i = 0 ;i < blocklengths.length; i++)
        for (j = 0; j < blocklengths[i]; j++)
          for (l = 0;l < oldtype.Size();l++){
            result.displacements[ptr] = oldtype.displacements[l] + 
                                        oldtype.Extent() +
                                        (_displacements[i] + j);

            if (result.displacements[ptr] > result.ub)
              result.ub = result.displacements[ptr];
            if (result.displacements[ptr] < result.lb)
              result.lb = result.displacements[ptr];
            ptr++;
          }
      result._extent = result.ub - result.lb + 1;

      return result;   
    }
    else
      return new Datatype(GetIndexed(blocklengths,_displacements,oldtype), 
                          oldtype);
  }

  private static native long GetIndexed(int[]      blocklengths, 
                                        int[]      displacements, 
                                        Datatype oldtype) ;

  /**
   * Identical to <tt>indexed</tt> except that the displacements are
   * expressed directly in terms of the buffer index, rather than the
   * units of the old type.
   * <p>
   * <table>
   * <tr><td><tt> array_of_blocklengths  </tt></td><td> number of elements per
   *                                                    block </tr>
   * <tr><td><tt> array_of_displacements </tt></td><td> displacement in buffer
   *                                                    for each block </tr>
   * <tr><td><tt> oldtype                </tt></td><td> old datatype </tr>
   * <tr><td><em> returns:               </em></td><td> new datatype </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_HINDEXED</tt>.
   * <p>
   * <em>Unlike other language bindings</em>, the values in
   * <tt>array_of_displacements</tt> are <em>not</em> measured in bytes.
   */

  public static Datatype Hindexed(int[]    blocklengths, 
                                  int[]    _displacements, 
                                  Datatype oldtype){
  
    if (oldtype.baseType == OBJECT){
      Datatype result = new Datatype(0,MPI.OBJECT);
      int ptr = 0, size = 0, i , j, l;
      result.lb = 2147483647;result.ub = -2147483647;

      for (i = 0; i < blocklengths.length; i++)
        size += blocklengths[i] * oldtype.Size();
  
      result.displacements = new int[size];

      for (i = 0 ;i < blocklengths.length; i++)
        for (j = 0; j < blocklengths[i]; j++)
          for (l = 0;l < oldtype.Size();l++){
            result.displacements[ptr] = oldtype.displacements[l] +
                                        _displacements[i] + 
                                        oldtype.Extent() * j;

            if (result.displacements[ptr] > result.ub)
              result.ub = result.displacements[ptr];
            if (result.displacements[ptr] < result.lb)
              result.lb = result.displacements[ptr];
            ptr++;
          }
      result._extent = result.ub - result.lb + 1;
      return result;  
    }
    else
      return new Datatype(GetHindexed(blocklengths,_displacements,oldtype), 
                          oldtype);
  }

  private static native long GetHindexed(int[]    blocklengths, 
                                         int[]    displacements,
                                         Datatype oldtype) ;

  /**
   * The most general type constructor.
   * <p>
   * <table>
   * <tr><td><tt> array_of_blocklengths  </tt></td><td> number of elements per
   *                                                    block </tr>
   * <tr><td><tt> array_of_displacements </tt></td><td> displacement in buffer
   *                                                    for each block </tr>
   * <tr><td><tt> array_of_types         </tt></td><td> type of elements in
   *                                                    each block </tr>
   * <tr><td><em> returns:               </em></td><td> new datatype </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TYPE_STRUCT</tt>.
   * <p>
   * The number of blocks is taken to be size of the
   * <tt>array_of_blocklengths</tt> argument.  The second and third
   * arguments, <tt>array_of_displacements</tt>, and <tt>array_of_types</tt>,
   * should be the same size.
   * <em>Unlike other language bindings</em>, the values in
   * <tt>array_of_displacements</tt> are <em>not</em> measured in bytes.
   * All elements of <tt>array_of_types</tt> with definite base types
   * <em>must have the <em>same</em> base type</em>: this will be the base
   * type of new datatype.
   */
  public static Datatype Struct(int[]      blocklengths,   
    				int[]      _displacements, 
    				Datatype[] types) {
  
    int baseType = UNDEFINED;
    for (int i = 0; i < types.length; i++)
      if (types[i].baseType != UNDEFINED && 
          types[i].baseType != UB &&
          types[i].baseType != LB) {

        if (baseType == UNDEFINED)
          baseType = types[i].baseType;
        else if (types[i].baseType != baseType){
	   System.out.println();
           System.out.println("Error ....");
           System.out.println("All type should be same or combine with UB, LB...");
           MPI.Finalize();
           System.exit(0);
        }
      }

    if (baseType == OBJECT){
      Datatype result = new Datatype(0,MPI.OBJECT);
      int size = 0,i,j,l;
      for (i = 0; i < blocklengths.length; i++)
        size += blocklengths[i] * types[i].Size();

      result.displacements = new int[size];
      result.lb = 2147483647;result.ub = -2147483647;
   
      int ptr = 0;

      for (i = 0;i < blocklengths.length;i++)
        for (j = 0; j < blocklengths[i] ; j++)  
          for (l = 0; l < types[i].Size(); l++){
            result.displacements[ptr] = types[i].displacements[l] + 
                                        _displacements[i] + 
                                        types[i].Extent()*j;

            if (result.displacements[ptr] > result.ub)
              result.ub = result.displacements[ptr];
            if (result.displacements[ptr] < result.lb)
              result.lb = result.displacements[ptr];
            ptr++;
          }
      result._extent = result.ub - result.lb + 1;   
      return result;
    }
    else 
       return new Datatype(GetStruct(blocklengths, _displacements, types), 
                           types[0]);
  }

  private static native long GetStruct(int[]      blocklengths, 
                                       int[]      displacements, 
                                       Datatype[] types) ;

  protected long handle;
  protected int  baseType;
  protected int displacements[]={0};
  protected int _extent = 1;
  protected int ub=0,lb=0;

  static {
    init();
  }

}

