/*
 * File         : Intracommm.java
 * Author       : Sang Lim, Xinying Li
 * Created      : Thu Apr  9 12:22:15 1998
 * Revision     : $Revision: 1.6 $
 * Updated      : $Date: 1999/03/05 18:02:15 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

package mpi;
//import mpi.*;

public class Intracomm extends Comm {

  public Intracomm() {}
  public Intracomm(int Type) { GetComm(Type);}
  public Intracomm(long _handle) { handle=_handle; }

  public Object clone() {
    return new Intracomm(dup()) ;
  }

  /**
   * Partition the group associated with this communicator and create
   * a new communicator within each subgroup.
   * <p>
   * <table>
   * <tr><td><tt> color    </tt></td><td> control of subset assignment </tr>
   * <tr><td><tt> key      </tt></td><td> control of rank assignment </tr>
   * <tr><td><em> returns: </em></td><td> new communicator </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_COMM_SPLIT</tt>.
   */

  public Intracomm Split(int colour, int key) {
    return new Intracomm(split(colour,key)) ;
  }

  private native long split(int colour, int key);

  /**
   * Create a new communicator.
   * <p>
   * <table>
   * <tr><td><tt> group    </tt></td><td> group which is a subset of the
   *                                      group of this communicator </tr>
   * <tr><td><em> returns: </em></td><td> new communicator </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_COMM_CREATE</tt>.
   */

  public Intracomm Creat(Group group) {
    long handle = creat(group) ;
    if(handle == 0) return null ;
    return new Intracomm(handle) ;
  }

  private native long creat(Group group);

  // Collective Communication

  /**
   * A call to <tt>Barrier</tt> blocks the caller until all process
   * in the group have called it.
   * <p>
   * Java binding of the MPI operation <tt>MPI_BARRIER</tt>.
   */

  public native void Barrier() ;

  /**
   * Broadcast a message from the process with rank <tt>root</tt>
   * to all processes of the group.
   * <p>
   * <table>
   * <tr><td><tt> buf      </tt></td><td> buffer array </tr>
   * <tr><td><tt> offset   </tt></td><td> initial offset in buffer </tr>
   * <tr><td><tt> count    </tt></td><td> number of items in buffer </tr>
   * <tr><td><tt> datatype </tt></td><td> datatype of each item in
   *                                      buffer </tr>
   * <tr><td><tt> root     </tt></td><td> rank of broadcast root </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_BCST</tt>.
   */

  public void Bcast(Object   buf, 
                    int      offset, 
                    int      count, 
                    Datatype type,
                    int      root){ 

    if (type.isObject(type)){
      byte[] byte_buf = Object_Serialize(buf,offset,count,type);
     
      bcast(byte_buf, 0,byte_buf.length,MPI.BYTE,root); 
       
      Object_Deserialize(buf,byte_buf,offset,count,type);
    }
    else
       bcast(buf,offset*type.Size(),count,type,root);   
  }

  private native void bcast(Object   buf, 
                            int      offset, 
                            int      count, 
                            Datatype type,
                            int      root);

  /**
   * Each process sends the contents of its send buffer to the
   * root process.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> sendcount  </tt></td><td> number of items to send </tr>
   * <tr><td><tt> sendtype   </tt></td><td> datatype of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcount  </tt></td><td> number of items to receive </tr>
   * <tr><td><tt> recvtype   </tt></td><td> datatype of each item in receive
   *                                        buffer </tr>
   * <tr><td><tt> root       </tt></td><td> rank of receiving process </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_GATHER</tt>.
   */

  public void Gather(Object   sendbuf, 
                     int      sendoffset, 
                     int      sendcount,
                     Datatype sendtype, 
                     Object   recvbuf, 
                     int      recvoffset, 
                     int      recvcount,
                     Datatype recvtype, 
                     int      root){
  
    if (sendtype.isObject(sendtype)){
      byte[] send_byte_buf = Object_Serialize(sendbuf,sendoffset,
                                              sendcount,sendtype);
      int i;
      int[] length_buf = new int[1];
      length_buf[0] = send_byte_buf.length; 

      if (MPI.COMM_WORLD.Rank() == root){
        int np = MPI.COMM_WORLD.Size();
        int [] recv_length_buf = new int[np];

        gather(length_buf,      0, length_buf.length, MPI.INT, 
               recv_length_buf, 0, length_buf.length, MPI.INT, 
               root);

        int buf_size = 0;
        for (i = 0; i < recv_length_buf.length;i++)
           buf_size += recv_length_buf[i];

        byte[] recv_byte_buf = new byte[buf_size];

        gather(send_byte_buf, 0, send_byte_buf.length, MPI.BYTE, 
               recv_byte_buf, 0, send_byte_buf.length, MPI.BYTE, 
               root);

        byte[] byte_buf = new byte[buf_size];

        int position = 0;
        int offset = recvoffset;
        for(i = 0; i < np; i++){
          System.arraycopy(recv_byte_buf,position,byte_buf,
                           0,recv_length_buf[i]);
          position += recv_length_buf[i];

          Object_Deserialize(recvbuf,byte_buf,offset,recvcount,recvtype); 

          offset += recvcount;
        }
      }
      else {
        int [] recv_length_buf = new int[1];
        byte[] recv_byte_buf = new byte[1];

        gather(length_buf,      0, length_buf.length, MPI.INT, 
               recv_length_buf, 0, length_buf.length, MPI.INT, 
               root);

        gather(send_byte_buf, 0, send_byte_buf.length, MPI.BYTE, 
               recv_byte_buf, 0, send_byte_buf.length, MPI.BYTE, 
               root);
      }
    }
    else
      gather(sendbuf, sendoffset*sendtype.Size(), sendcount,sendtype,
	     recvbuf, recvoffset*recvtype.Size(), recvcount,recvtype,
             root);
  }

  private native void gather(Object   sendbuf, 
                             int      sendoffset, 
                             int      sendcount,
	                     Datatype sendtype, 
                             Object   recvbuf, 
                             int      recvoffset, 
                             int      recvcount,
	                     Datatype recvtype, 
                             int      root);

  /**
   * Extends functionality of <tt>Gather</tt> by allowing varying
   * counts of data from each process.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> sendcount  </tt></td><td> number of items to send </tr>
   * <tr><td><tt> sendtype   </tt></td><td> datatype of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcounts </tt></td><td> number of elements received from
   *                                        each process </tr>
   * <tr><td><tt> displs     </tt></td><td> displacements at which to place
   *                                        incoming data </tr>
   * <tr><td><tt> recvtype   </tt></td><td> datatype of each item in receive
   *                                        buffer </tr>
   * <tr><td><tt> root       </tt></td><td> rank of receiving process </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_GATHERV</tt>.
   * <p>
   * The sizes of arrays <tt>recvcounts</tt> and <tt>displs</tt> should be the
   * size of the group.  Entry <em>i</em> of <tt>displs</tt> specifies the
   * displacement relative to element <tt>recvoffset</tt> of <tt>recvbuf</tt>
   * at which to place incoming data.
   */

  public void Gatherv(Object   sendbuf,  
                      int      sendoffset, 
                      int      sendcount,
                      Datatype sendtype, 
                      Object   recvbuf, 
                      int      recvoffset, 
                      int []   recvcount,
                      int []   displs, 
                      Datatype recvtype, 
                      int      root){

    if (sendtype.isObject(sendtype)){
      int np = MPI.COMM_WORLD.Size();
      byte[] send_byte_buf = Object_Serialize(sendbuf,sendoffset,
                                              sendcount,sendtype);
      int i;
      int[] length_buf = new int[1];
      length_buf[0] = send_byte_buf.length; 

      int rcount[] = new int[np];
      int dis[] = new int[np];       

      if (MPI.COMM_WORLD.Rank() == root){
        int [] recv_length_buf = new int[np];

        gather(length_buf,      0, length_buf.length, MPI.INT, 
               recv_length_buf, 0, length_buf.length, MPI.INT, 
               root);

        int buf_size = 0;
        int count = 0;
        for (i = 0; i < recv_length_buf.length;i++){
          buf_size += recv_length_buf[i];
          rcount[i] = recv_length_buf[i];
        }

        dis[0] = 0;
        for (i = 0; i < recv_length_buf.length-1; i++)
           dis[i+1] = dis[i] + recv_length_buf[i];
   
        byte[] recv_byte_buf = new byte[buf_size];

        gatherv(send_byte_buf, 0, send_byte_buf.length, MPI.BYTE, 
                recv_byte_buf, 0, rcount, dis         , MPI.BYTE, 
                root);

        byte[] byte_buf = new byte[buf_size];

        int position = 0;
        int offset = 0;
        for(i = 0; i < np; i++){
          System.arraycopy(recv_byte_buf,position,byte_buf,
                           0,recv_length_buf[i]);
          position += recv_length_buf[i];
          
          offset = displs[i] + recvoffset;
          Object_Deserialize(recvbuf,byte_buf,offset,recvcount[i],recvtype); 
        }
      }
      else {
        int [] recv_length_buf = new int[1];
        byte[] recv_byte_buf = new byte[1];

        gather(length_buf,      0, length_buf.length, MPI.INT, 
               recv_length_buf, 0, length_buf.length, MPI.INT, 
               root);

        gatherv(send_byte_buf, 0, send_byte_buf.length, MPI.BYTE, 
                recv_byte_buf, 0, rcount, dis         , MPI.BYTE, 
                root);
      }   

    }
    else
      gatherv(sendbuf  , sendoffset*sendtype.Size(), 
              sendcount, sendtype,
              recvbuf  , recvoffset*recvtype.Size(),
              recvcount, displs,
              recvtype , root);
  }

  private native void gatherv(Object   sendbuf, 
                              int      sendoffset, 
                              int      sendcount,
  	                      Datatype sendtype, 
                              Object   recvbuf, 
                              int      recvoffset, 
                              int []   recvcount,
 	                      int []   displs, 
                              Datatype recvtype, 
                              int      root);

  /**
   * Inverse of the operation <tt>Gather</tt>.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> sendcount  </tt></td><td> number of items to send </tr>
   * <tr><td><tt> sendtype   </tt></td><td> datatype of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcount  </tt></td><td> number of items to receive </tr>
   * <tr><td><tt> recvtype   </tt></td><td> datatype of each item in receive
   *                                        buffer </tr>
   * <tr><td><tt> root       </tt></td><td> rank of sending process </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_SCATTER</tt>.
   */

  public void Scatter(Object   sendbuf, 
                      int      sendoffset, 
                      int      sendcount,
                      Datatype sendtype, 
                      Object   recvbuf, 
                      int      recvoffset, 
                      int      recvcount,
                      Datatype recvtype, 
                      int      root){

    if (sendtype.isObject(sendtype)){
      int i;

      if (MPI.COMM_WORLD.Rank() == root){
        int np = MPI.COMM_WORLD.Size();
        for (i = 0; i < np ;i++){
          byte[] send_byte_buf;

          send_byte_buf = Object_Serialize(sendbuf, sendoffset+sendcount*i,
                                           sendcount, sendtype);

          int[] length_buf = new int[1];
          length_buf[0] = send_byte_buf.length; 
          Send(length_buf, 0, 1, MPI.INT, i, 1);

          Send(send_byte_buf, 0, length_buf[0],
               MPI.BYTE,      i, 1);
        }
      }
        
      int[] length_buf= new int[1];
      Recv(length_buf, 0, 1, MPI.INT, root, 1); 

      byte[] byte_buf = new byte[length_buf[0]];
      Recv(byte_buf, 0, length_buf[0], MPI.BYTE, root, 1); 

      Object_Deserialize(recvbuf,byte_buf,recvoffset,recvcount,recvtype);

    }
    else 
      scatter(sendbuf,   sendoffset*sendtype.Size(), 
              sendcount, sendtype, 
              recvbuf,   recvoffset*recvtype.Size(), 
              recvcount, recvtype, 
              root);
  }

  private native void scatter(Object   sendbuf, 
                              int      sendoffset, 
                              int      sendcount,
                              Datatype sendtype, 
                              Object   recvbuf, 
                              int      recvoffset, 
                              int      recvcount,
                              Datatype recvtype, 
                              int      root);

  /**
   * Inverse of the operation <tt>Gatherv</tt>.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> sendcounts </tt></td><td> number of items sent to each
   *                                        process </tr>
   * <tr><td><tt> displs     </tt></td><td> displacements from which to take
   *                                        outgoing data </tr>
   * <tr><td><tt> sendtype   </tt></td><td> datatype of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcount  </tt></td><td> number of items to receive </tr>
   * <tr><td><tt> recvtype   </tt></td><td> datatype of each item in receive
   *                                        buffer </tr>
   * <tr><td><tt> root       </tt></td><td> rank of sending process </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_SCATTERV</tt>.
   */

  public void Scatterv(Object   sendbuf, 
                       int      sendoffset,
                       int []   sendcount,
                       int []   displs,
                       Datatype sendtype, 
                       Object   recvbuf, 
                       int      recvoffset, 
                       int      recvcount,
                       Datatype recvtype, 
                       int      root){
  
    if (sendtype.isObject(sendtype)){
      int i;

      if (MPI.COMM_WORLD.Rank() == root){
        int np = MPI.COMM_WORLD.Size();

        for (i = 0; i < np ;i++){
          byte[] send_byte_buf;

          send_byte_buf = Object_Serialize(sendbuf, sendoffset+displs[i],
                                           sendcount[i], sendtype);

          int[] length_buf = new int[1];
          length_buf[0] = send_byte_buf.length; 
          Send(length_buf, 0, 1, MPI.INT, i, 1);

          Send(send_byte_buf, 0, length_buf[0],
               MPI.BYTE,      i, 1);
        }
      }
        
      int[] length_buf= new int[1];
      Recv(length_buf, 0, 1, MPI.INT, root, 1); 

      byte[] byte_buf = new byte[length_buf[0]];
      Recv(byte_buf, 0, length_buf[0], MPI.BYTE, root, 1); 

      Object_Deserialize(recvbuf,byte_buf,recvoffset,recvcount,recvtype);

    }
    else
      scatterv(sendbuf                   , sendoffset*sendtype.Size(), 
	       sendcount                 , displs, 
               sendtype                  , recvbuf, 
               recvoffset*recvtype.Size(),  recvcount,
               recvtype                  ,  root);
  }

  private native void scatterv(Object   sendbuf, 
                               int      sendoffset, 
          	               int []   sendcount,
                               int []   displs,
                               Datatype sendtype, 
                               Object   recvbuf, 
                               int      recvoffset, 
                               int      recvcount,
                               Datatype recvtype, 
                               int      root);

  /**
   * Similar to <tt>Gather</tt>, but all processes receive the result.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> sendcount  </tt></td><td> number of items to send </tr>
   * <tr><td><tt> sendtype   </tt></td><td> datatype of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcount  </tt></td><td> number of items to receive </tr>
   * <tr><td><tt> recvtype   </tt></td><td> datatype of each item in receive
   *                                        buffer </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_ALLGATHER</tt>.
   */

  public void Allgather(Object   sendbuf, 
                        int      sendoffset, 
                        int      sendcount,
                        Datatype sendtype, 
                        Object   recvbuf, 
                        int      recvoffset, 
                        int      recvcount,
                        Datatype recvtype){
 
    if (sendtype.isObject(sendtype)){
      int i;
      byte[] send_byte_buf = Object_Serialize(sendbuf,sendoffset,
                                              sendcount,sendtype);

      int[] send_length_buf = new int[1];
      int np = MPI.COMM_WORLD.Size();
      int[] recv_length_buf = new int[np];

      send_length_buf[0] = send_byte_buf.length;

      allgather(send_length_buf, 0, send_length_buf.length,MPI.INT, 
                recv_length_buf, 0, send_length_buf.length,MPI.INT);

      int buf_size = 0,count = 0;
      for (i = 0; i < recv_length_buf.length;i++)
        buf_size += recv_length_buf[i];

      byte[] recv_byte_buf = new byte[buf_size];

      allgather(send_byte_buf, 0, send_byte_buf.length,MPI.BYTE, 
                recv_byte_buf, 0, send_byte_buf.length,MPI.BYTE);

      byte[] byte_buf = new byte[buf_size];

      int position = 0;
      int offset = recvoffset;
      for(i = 0; i < np; i++){
        System.arraycopy(recv_byte_buf,position,byte_buf,
                         0,recv_length_buf[i]);
        position += recv_length_buf[i];

        Object_Deserialize(recvbuf,byte_buf,offset,recvcount,recvtype); 
        offset += recvcount;
      }
    }
    else 
      allgather(sendbuf,   sendoffset*sendtype.Size(), 
                sendcount, sendtype, 
                recvbuf,   recvoffset*recvtype.Size(), 
                recvcount, recvtype);
  }

  private native void allgather(Object   sendbuf, 
                                int      sendoffset, 
                                int      sendcount,
                                Datatype sendtype, 
                                Object   recvbuf, 
                                int      recvoffset, 
                                int      recvcount,
                                Datatype recvtype);


  /**
   * Similar to <tt>Gatherv</tt>, but all processes receive the result.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> sendcount  </tt></td><td> number of items to send </tr>
   * <tr><td><tt> sendtype   </tt></td><td> datatype of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcounts </tt></td><td> number of elements received from
   *                                        each process </tr>
   * <tr><td><tt> displs     </tt></td><td> displacements at which to place
   *                                        incoming data </tr>
   * <tr><td><tt> recvtype   </tt></td><td> datatype of each item in receive
   *                                        buffer </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_ALLGATHERV</tt>.
   */

  public void Allgatherv(Object   sendbuf, 
                         int      sendoffset, 
                         int      sendcount,
                         Datatype sendtype, 
                         Object   recvbuf, 
                         int      recvoffset, 
                         int []   recvcount,
                         int []   displs, 
                         Datatype recvtype){

    if (sendtype.isObject(sendtype)){
      int i;
      int np = MPI.COMM_WORLD.Size();

      byte[] send_byte_buf = Object_Serialize(sendbuf,sendoffset,
                                              sendcount,sendtype);

      int[] send_length_buf = new int[1];
      int[] recv_length_buf = new int[np];

      send_length_buf[0] = send_byte_buf.length;
      int rcount[] = new int[np];
      int dis[] = new int[np];

      allgather(send_length_buf, 0, send_length_buf.length,MPI.INT, 
                recv_length_buf, 0, send_length_buf.length,MPI.INT);

      int buf_size = 0;
      for (i = 0; i < recv_length_buf.length;i++){
        buf_size += recv_length_buf[i];
        rcount[i] = recv_length_buf[i];
      }
      dis[0] = 0;
      for(i = 0; i < recv_length_buf.length-1; i++)
        dis[i+1] = dis[i] + recv_length_buf[i];

      byte[] recv_byte_buf = new byte[buf_size];

      allgatherv(send_byte_buf, 0, send_byte_buf.length, MPI.BYTE, 
                 recv_byte_buf, 0, rcount, dis         , MPI.BYTE);

      byte[] byte_buf = new byte[buf_size];

      int j=0,k=0;
      int init = recv_length_buf[k];

      int position = 0;
      int offset = 0;
      for(i = 0; i < np; i++){
        System.arraycopy(recv_byte_buf,position,byte_buf,
                         0,recv_length_buf[i]);
        position += recv_length_buf[i];
          
        offset = displs[i] + recvoffset;
        Object_Deserialize(recvbuf,byte_buf,offset,recvcount[i],recvtype); 
      }
    }
    else
      allgatherv(sendbuf  , sendoffset*sendtype.Size(), 
                 sendcount, sendtype, 
                 recvbuf  , recvoffset*recvtype.Size(), 
                 recvcount, displs, 
                 recvtype);
  }

  private native void allgatherv(Object   sendbuf, 
                                 int      sendoffset, 
                                 int      sendcount,
                                 Datatype sendtype, 
                                 Object   recvbuf, 
                                 int      recvoffset, 
                                 int []   recvcount,
                                 int []   displs, 
                                 Datatype recvtype);

  /**
   * Extension of <tt>Allgather</tt> to the case where each process sends
   * distinct data to each of the receivers.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> sendcount  </tt></td><td> number of items sent to each
   *                                        process </tr>
   * <tr><td><tt> sendtype   </tt></td><td> datatype send buffer items </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcount  </tt></td><td> number of items received from any
   *                                        process
   * <tr><td><tt> recvtype   </tt></td><td> datatype of receive buffer
   *                                        items </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_ALLTOALL</tt>.
   */

  public void Alltoall(Object   sendbuf, 
                       int      sendoffset, 
                       int      sendcount,
                       Datatype sendtype, 
                       Object   recvbuf, 
                       int      recvoffset, 
                       int      recvcount,
                       Datatype recvtype){

    if (sendtype.isObject(sendtype)){
      int i;
      int offset = recvoffset;
      for(i =0; i < MPI.COMM_WORLD.Size(); i++){
        for (int j = 0; j < MPI.COMM_WORLD.Size(); j++){
          if (i == MPI.COMM_WORLD.Rank()){
            byte[] send_byte_buf = Object_Serialize(sendbuf, 
                                                    sendoffset+sendcount*i,
                                                    sendcount, sendtype);
            int[] length_buf = new int[1];
            length_buf[0] = send_byte_buf.length; 
            Send(length_buf, 0, 1, MPI.INT, j, 1);

            Send(send_byte_buf, 0, length_buf[0],
                 MPI.BYTE,      j, 1);
          }
        }
        int[] length_buf= new int[1];
        Recv(length_buf, 0, 1, MPI.INT, i, 1); 

        byte[] recv_byte_buf = new byte[length_buf[0]];
        Recv(recv_byte_buf, 0, length_buf[0], MPI.BYTE, i, 1); 

        Object_Deserialize(recvbuf,recv_byte_buf,offset,
                           recvcount,recvtype);
        offset += recvcount;

      }              
   }
   else 
     alltoall(sendbuf,   sendoffset*sendtype.Size(), 
              sendcount, sendtype, 
              recvbuf,   recvoffset*recvtype.Size(), 
              recvcount, recvtype);
  }

  private native void alltoall(Object   sendbuf, 
                               int      sendoffset, 
                               int      sendcount,
                               Datatype sendtype, 
                               Object   recvbuf, 
                               int      recvoffset, 
                               int      recvcount,
                               Datatype recvtype);

  /**
   * Adds flexibility to <tt>Alltoall</tt>: location of data for send is
   * specified by <tt>sdispls</tt> and location to place data on receive
   * side is specified by <tt>rdispls</tt>.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> sendcounts </tt></td><td> number of items sent to each
   *                                        process </tr>
   * <tr><td><tt> sdispls    </tt></td><td> displacements from which to take
   *                                        outgoing data </tr>
   * <tr><td><tt> sendtype   </tt></td><td> datatype send buffer items </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcounts </tt></td><td> number of elements received from
   *                                        each process
   * <tr><td><tt> rdispls    </tt></td><td> displacements at which to place
   *                                        incoming data </tr>
   * <tr><td><tt> recvtype   </tt></td><td> datatype of each item in receive
   *                                        buffer </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_ALLTOALLV</tt>.
   */

  public void Alltoallv(Object   sendbuf, 
                        int      sendoffset, 
                        int []   sendcount,
                        int []   sdispls,
                        Datatype sendtype, 
                        Object   recvbuf, 
                        int      recvoffset, 
                        int []   recvcount,
                        int []   rdispls, 
                        Datatype recvtype){

    if (sendtype.isObject(sendtype)){
      int count = 0;
      for(int j = 0; j < sdispls.length;j++)
        count += sdispls[j];

      byte[] temp_buf = Object_Serialize(sendbuf, 0,
                                         count, 
                                         sendtype);

      int offset = 0;
      for(int i =0; i < MPI.COMM_WORLD.Size(); i++){
        for (int j = 0; j < MPI.COMM_WORLD.Size(); j++){
          if (i == MPI.COMM_WORLD.Rank()){
             byte[] send_byte_buf = Object_Serialize(sendbuf, 
                                                     sendoffset+sdispls[i],
                                                     sendcount[i], sendtype);
             int[] length_buf = new int[1];
             length_buf[0] = send_byte_buf.length; 
             Send(length_buf, 0, 1, MPI.INT, j, 1);

             Send(send_byte_buf, 0, length_buf[0],
                  MPI.BYTE,      j, 1);
           }
         }
         int[] length_buf= new int[1];
         Recv(length_buf, 0, 1, MPI.INT, i, 1); 

         byte[] recv_byte_buf = new byte[length_buf[0]];
         Recv(recv_byte_buf, 0, length_buf[0], MPI.BYTE, i, 1); 

         offset = rdispls[i] + recvoffset;
         Object_Deserialize(recvbuf,recv_byte_buf,offset,recvcount[i],recvtype); 
      }
     }
     else 
       alltoallv(sendbuf,   sendoffset*sendtype.Size(), 
                 sendcount, sdispls, sendtype, 
                 recvbuf,   recvoffset*recvtype.Size(), 
                 recvcount, rdispls, recvtype);
  }

  private native void alltoallv(Object   sendbuf, 
                                int      sendoffset, 
                                int []   sendcount,
           	                int []   sdispls,
                                Datatype sendtype, 
                                Object   recvbuf, 
                                int      recvoffset, 
                                int []   recvcount,
                                int []   displs, 
                                Datatype recvtype);

  /**
   * Combine elements in input buffer of each process using the reduce
   * operation, and return the combined value in the output buffer of the
   * root process.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> count      </tt></td><td> number of items in send buffer </tr>
   * <tr><td><tt> datatype   </tt></td><td> data type of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> op         </tt></td><td> reduce operation </tr>
   * <tr><td><tt> root       </tt></td><td> rank of root process </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_REDUCE</tt>.
   * <p>
   * The predefined operations are available in Java as <tt>MPI.MAX</tt>,
   * <tt>MPI.MIN</tt>, <tt>MPI.SUM</tt>, <tt>MPI.PROD</tt>, <tt>MPI.LAND</tt>,
   * <tt>MPI.BAND</tt>, <tt>MPI.LOR</tt>, <tt>MPI.BOR</tt>, <tt>MPI.LXOR</tt>,
   * <tt>MPI.BXOR</tt>, <tt>MPI.MINLOC</tt> and <tt>MPI.MAXLOC</tt>.
   */

  public void Reduce(Object sendbuf, int sendoffset,
                     Object recvbuf, int recvoffset, int count,
                     Datatype datatype, Op op, int root) {
    reduce(sendbuf, sendoffset * datatype.baseSize(), 
           recvbuf, recvoffset * datatype.baseSize(), count,
           datatype, op, root) ;
  }

  private native void reduce(Object sendbuf, int sendoffset, 
                             Object recvbuf, int recvoffset, int count,
                             Datatype datatype, Op op, int root);

  /**
   * Same as <tt>reduce</tt> except that the result appears in receive
   * buffer of all process in the group.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> count      </tt></td><td> number of items in send buffer </tr>
   * <tr><td><tt> datatype   </tt></td><td> data type of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> op         </tt></td><td> reduce operation </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_ALLREDUCE</tt>.
   */

  public void Allreduce(Object sendbuf, int sendoffset,
                        Object recvbuf, int recvoffset, int count,
                        Datatype datatype, Op op) {
    allreduce(sendbuf, sendoffset * datatype.baseSize(), 
              recvbuf, recvoffset * datatype.baseSize(), count, datatype, op) ;
  }

  private native void allreduce(Object sendbuf, int sendoffset, 
                                Object recvbuf, int recvoffset, int count,
                                Datatype datatype, Op op) ;

  /**
   * Combine elements in input buffer of each process using the reduce
   * operation, and scatter the combined values over the output buffers
   * of the processes.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> recvcounts </tt></td><td> numbers of result elements
   *                                        distributed to each process </tr>
   * <tr><td><tt> datatype   </tt></td><td> data type of each item in send
   *                                        buffer </tr>
   * <tr><td><tt> op         </tt></td><td> reduce operation </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_REDUCE_SCATTER</tt>.
   */

  public void Reduce_scatter(Object sendbuf, int sendoffset,
                             Object recvbuf, int recvoffset, int [] recvcounts,
                             Datatype datatype, Op op) {
    reduce_scatter(sendbuf, sendoffset * datatype.baseSize(),
                   recvbuf, recvoffset * datatype.baseSize(), recvcounts, 
	           datatype, op) ;
  }

  private native void reduce_scatter(Object sendbuf, int sendoffset, 
                                     Object recvbuf, int recvoffset,
                                     int [] recvcounts, 
	                             Datatype datatype, Op op) ;

  /**
   * Perform a prefix reduction on data distributed across the group.
   * <p>
   * <table>
   * <tr><td><tt> sendbuf    </tt></td><td> send buffer array </tr>
   * <tr><td><tt> sendoffset </tt></td><td> initial offset in send buffer </tr>
   * <tr><td><tt> recvbuf    </tt></td><td> receive buffer array </tr>
   * <tr><td><tt> recvoffset </tt></td><td> initial offset in receive
   *                                        buffer </tr>
   * <tr><td><tt> count      </tt></td><td> number of items in input
   *                                        buffer </tr>
   * <tr><td><tt> datatype   </tt></td><td> data type of each item in input
   *                                        buffer </tr>
   * <tr><td><tt> op         </tt></td><td> reduce operation </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_SCAN</tt>.
   */

  public void Scan(Object sendbuf, int sendoffset,
                   Object recvbuf, int recvoffset, int count,
                   Datatype datatype, Op op) {
    scan(sendbuf, sendoffset * datatype.baseSize(), 
         recvbuf, recvoffset * datatype.baseSize(), count, datatype, op);
  }

  private native void scan(Object sendbuf, int sendoffset, 
                           Object recvbuf, int recvoffset, int count,
                           Datatype datatype, Op op) ;

  // Topology Constructors

  /**
   * Create a Cartesian topology communicator whose group is a subset
   * of the group of this communicator.
   * <p>
   * <table>
   * <tr><td><tt> dims     </tt></td><td> the number of processes in each
   *                                      dimension </tr>
   * <tr><td><tt> periods  </tt></td><td> <tt>true</tt> if grid is periodic,
   *                                      <tt>false</tt> if not, in each
   *                                      dimension </tr>
   * <tr><td><tt> reorder  </tt></td><td> <tt>true</tt> if ranking may be
   *                                      reordered, <tt>false</tt> if not </tr>
   * <tr><td><em> returns: </em></td><td> new Cartesian topology
   *                                      communicator </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_CART_CREATE</tt>.
   * <p>
   * The number of dimensions of the Cartesian grid is taken to be the size
   * of the <tt>dims</tt> argument.  The array <tt>periods</tt> must be the
   * same size.
   */

  public Cartcomm Create_cart(int [] dims, boolean [] periods, 
	                      boolean reorder) {
    return new Cartcomm(GetCart(dims, periods, reorder)) ;
  }

  private native long GetCart(int [] dims, boolean [] periods,
                              boolean reorder) ;

  /**
   * Create a graph topology communicator whose group is a subset
   * of the group of this communicator.
   * <p>
   * <table>
   * <tr><td><tt> index    </tt></td><td> node degrees </tr>
   * <tr><td><tt> edges    </tt></td><td> graph edges </tr>
   * <tr><td><tt> reorder  </tt></td><td> <tt>true</tt> if ranking may be
   *                                      reordered, <tt>false</tt> if not </tr>
   * <tr><td><em> returns: </em></td><td> new graph topology communicator </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_GRAPH_CREATE</tt>.
   * <p>
   * The number of nodes in the graph, <em>nnodes</em>, is taken
   * to be size of the <tt>index</tt> argument.  The size of array
   * <tt>edges</tt> must be <tt>index [nnodes} - 1]</tt>.
   */

  public Graphcomm Create_graph(int [] index, int [] edges, boolean reorder) {
    return new Graphcomm(GetGraph(index,edges,reorder)) ;
  }

  private native long GetGraph(int [] index,int [] edges, boolean reorder);
}












