/*
 * File         : Request.java
 * Author       : Sang Lim, Xinying Li
 * Created      : Thu Apr  9 12:22:15 1998
 * Revision     : $Revision: 1.4 $
 * Updated      : $Date: 1999/03/03 19:55:43 $
 * Copyright: Northeast Parallel Architectures Center
 *            at Syracuse University 1998
 */

package mpi;
//import mpi.*;

public class Request {
  public final static int NULL = 0;
  public final static int Tag_Normal = 0;
  public final static int Tag_Isend_Object = 1;
  public final static int Tag_Irecv_Object = 2;
  public final static int Tag_Send_Object  = 3;
  public final static int Tag_Bsend_Object = 4;
  public final static int Tag_Ssend_Object = 5;
  public final static int Tag_Rsend_Object = 6;
  public final static int Tag_Recv_Object  = 7;

  public Request dataReq;
  public int opTag = Tag_Normal;
  public Object buf;
  public int[] length_buf;
  public byte[] byte_buf;
  public int offset;
  public int count;
  public Datatype type;
  public int dest;
  public int src;
  public int tag;
  public Comm comm;


  private static native void init();
 
  protected native void GetReq(int Type);


  public Request() {}
  public Request(int Type) { GetReq(Type); }
  public Request(long _handle) { handle = _handle;}

  /**
   * Set the request object to be void.
   * Java binding of the MPI operation <tt>MPI_REQUEST_FREE</tt>.
   */

  public native void Free();

  /**
   * Mark a pending nonblocking communication for cancellation.
   * Java binding of the MPI operation <em>MPI_CANCEL</em>.
   */

  public native void Cancel();

  // Nonblocking communication

  /**
   * Test if request object is void.
   * <p>
   * <table>
   * <tr><td><em> returns: </em></td><td> true if the request object is void,
   *                                      false otherwise </tr>
   * </table>
   */

  public native boolean Is_null();

  /**
   * Blocks until the operation identified by the request is complete.
   * <p>
   * <table>
   * <tr><td><em> returns: </em></td><td> status object </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_WAIT</tt>.
   * <p>
   * After the call returns, the request object becomes inactive.
   */

  private native Status Wait(Status stat);
  public Status Wait() {
    if (opTag == Tag_Normal)
      return Wait(new Status());   
     
    Status result = new Status();
    Wait(result);

    switch (opTag) {
      case Tag_Isend_Object :
          dataReq.Wait(new Status());
        break;
      case Tag_Irecv_Object :
          byte_buf = new byte[length_buf[0]];             
          comm.Recv(byte_buf,0,length_buf[0], MPI.BYTE, src, tag); 
          result.object_count = length_buf[1];  
          comm.Object_Deserialize(buf,byte_buf,offset,length_buf[1],type);
        break;
      case Tag_Send_Object :
          dataReq.Wait(new Status());
        break;
      case Tag_Bsend_Object :
          dataReq.Wait(new Status());
        break;
      case Tag_Ssend_Object :
          dataReq.Wait(new Status());
         break;
      case Tag_Rsend_Object :
          dataReq.Wait(new Status());
        break;
      case Tag_Recv_Object :
          comm.Recv(length_buf,0,2,MPI.INT, src, tag);

          byte_buf = new byte[length_buf[0]]; 
          comm.Recv(byte_buf,0, length_buf[0], MPI.BYTE,src,tag);
	  result.object_count = length_buf[1];

          comm.Object_Deserialize(buf,byte_buf,offset,length_buf[1],type);
        break; 
    }
    return result;
  }

  /**
   * Returns a status object if the operation identified by the request
   * is complete, or a null handle otherwise.
   * <p>
   * <table>
   * <tr><td><em> returns: </em></td><td> status object or null handle </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TEST</tt>.
   * <p>
   * After the call, if the operation is complete (ie, if the return value
   * is non-null), the request object becomes inactive.
   */

  private native Status Test(Status stat);
  public Status Test() {
    if (opTag == Tag_Normal)
      return Test(new Status()); 

    Status result = new Status();
    if (Test(result) == null)
      return null;
          
    switch (opTag) {
      case Tag_Isend_Object :
          dataReq.Wait(new Status());
        break;
      case  Tag_Irecv_Object :
          byte_buf = new byte[length_buf[0]];             
          comm.Recv(byte_buf,0,length_buf[0], MPI.BYTE, src, tag); 
          result.object_count = length_buf[1];  
          comm.Object_Deserialize(buf,byte_buf,offset,length_buf[1],type);
        break;
      case Tag_Send_Object :
          dataReq.Wait(new Status());
        break;
     case Tag_Bsend_Object :
         dataReq.Wait(new Status());
        break;
      case Tag_Ssend_Object :
          dataReq.Wait(new Status());
        break;
      case Tag_Rsend_Object :
          dataReq.Wait(new Status());
        break;
      case Tag_Recv_Object :
          comm.Recv(length_buf,0,2,MPI.INT, src, tag);

          byte_buf = new byte[length_buf[0]]; 
          comm.Recv(byte_buf,0, length_buf[0], MPI.BYTE,src,tag);
	  result.object_count = length_buf[1];

          comm.Object_Deserialize(buf,byte_buf,offset,length_buf[1],type);
        break; 
    }
    return result;
  }

  /**
   * Blocks until one of the operations associated with the active
   * requests in the array has completed.
   * <p>
   * <table>
   * <tr><td><tt> array_of_requests </tt></td><td> array of requests </tr>
   * <tr><td><em> returns:          </em></td><td> status object </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_WAITANY</tt>.
   * <p>
   * The index in <tt>array_of_requests</tt> for the request that completed
   * can be obtained from the returned status object through the
   * <tt>Status.index</tt> field.  The corresponding element
   * of <tt>array_of_requests</tt> becomes inactive.
   */

  private static native Status Waitany(Request [] array_of_request, 
                                       Status     stat);
  public static Status Waitany(Request [] array_of_request) {
    Status result = Waitany(array_of_request, new Status());
    
    if (result == null)
        return null;        
    
    int index = result.index;
   
    if (array_of_request[index].opTag == Tag_Normal)
       return result;   
       
    switch (array_of_request[index].opTag) {
      case Tag_Isend_Object :
          array_of_request[index].dataReq.Wait(new Status());
        break;
      case  Tag_Irecv_Object :
          byte[] byte_buf = new byte[array_of_request[index].length_buf[0]]; 
          result.object_count = array_of_request[index].length_buf[1];
 
          array_of_request[index].comm.Recv(byte_buf,0,
                   array_of_request[index].length_buf[0], MPI.BYTE,
                   array_of_request[index].src, array_of_request[index].tag);
            
          array_of_request[index].comm.Object_Deserialize(
                 array_of_request[index].buf,byte_buf,
                 array_of_request[index].offset,
                 array_of_request[index].length_buf[1],
                 array_of_request[index].type);
        break;
      case Tag_Send_Object :
          array_of_request[index].dataReq.Wait(new Status());
        break;
      case Tag_Bsend_Object :
          array_of_request[index].dataReq.Wait(new Status());
        break;
      case Tag_Ssend_Object :
          array_of_request[index].dataReq.Wait(new Status());
        break;
      case Tag_Rsend_Object :
          array_of_request[index].dataReq.Wait(new Status());
        break;
      case Tag_Recv_Object :
          array_of_request[index].comm.Recv(
                 array_of_request[index].length_buf,0,2,MPI.INT, 
                 array_of_request[index].src, 
                 array_of_request[index].tag);

          byte_buf = new byte[array_of_request[index].length_buf[0]]; 
          result.object_count = array_of_request[index].length_buf[1];
          array_of_request[index].comm.Recv(byte_buf,0, 
                 array_of_request[index].length_buf[0], MPI.BYTE,
                 array_of_request[index].src,array_of_request[index].tag);

          array_of_request[index].comm.Object_Deserialize(
                 array_of_request[index].buf,byte_buf,
                 array_of_request[index].offset,
                 array_of_request[index].length_buf[1],
                 array_of_request[index].type);
        break; 
    }
    return result;
  }

  /**
   * Tests for completion of either one or none of the operations associated
   * with active requests.
   * <p>
   * <table>
   * <tr><td><tt> array_of_requests </tt></td><td> array of requests </tr>
   * <tr><td><em> returns:          </em></td><td> status object or
   *                                               null handle </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TESTANY</tt>.
   * <p>
   * If some request completed, the index in <tt>array_of_requests</tt>
   * for that request can be obtained from the returned status object
   * through the.  The corresponding element of <tt>array_of_requests</tt>
   * becomes inactive.
   * If no request completed, <tt>Testany</tt> returns a null handle.
   */

  private static native Status Testany(Request [] array_of_request, 
                                       Status     stat);
  public static Status Testany(Request [] array_of_request) {
 
    Status result = Testany(array_of_request, new Status());
    if (result == null)
      return null;

    int index = result.index;
   
    if (array_of_request[index].opTag == Tag_Normal)
      return result;   
       
    switch (array_of_request[index].opTag) {
      case Tag_Isend_Object :
          array_of_request[index].dataReq.Test(new Status());
        break;
      case  Tag_Irecv_Object :
          byte[] byte_buf = new byte[array_of_request[index].length_buf[0]]; 
          result.object_count = array_of_request[index].length_buf[1];
 
          array_of_request[index].comm.Recv(byte_buf,0,
                 array_of_request[index].length_buf[0], MPI.BYTE,
                 array_of_request[index].src, array_of_request[index].tag);
            
          array_of_request[index].comm.Object_Deserialize(
                 array_of_request[index].buf,byte_buf,
                 array_of_request[index].offset,
                 array_of_request[index].length_buf[1],
                 array_of_request[index].type);
        break;
      case Tag_Send_Object :
          array_of_request[index].dataReq.Test(new Status());
        break;
      case Tag_Bsend_Object :
          array_of_request[index].dataReq.Test(new Status());
        break;
      case Tag_Ssend_Object :
          array_of_request[index].dataReq.Test(new Status());
        break;
      case Tag_Rsend_Object :
          array_of_request[index].dataReq.Test(new Status());
        break;
      case Tag_Recv_Object :
          array_of_request[index].comm.Recv(
                 array_of_request[index].length_buf,0,2,MPI.INT, 
                 array_of_request[index].src, 
                 array_of_request[index].tag);

          byte_buf = new byte[array_of_request[index].length_buf[0]]; 
          result.object_count = array_of_request[index].length_buf[1];
          array_of_request[index].comm.Recv(byte_buf,0, 
                 array_of_request[index].length_buf[0], MPI.BYTE,
                 array_of_request[index].src,array_of_request[index].tag);

          array_of_request[index].comm.Object_Deserialize(
                 array_of_request[index].buf,byte_buf,
                 array_of_request[index].offset,
                 array_of_request[index].length_buf[1],
                 array_of_request[index].type);
        break; 
    }
    return result;
  }

  /**
   * Blocks until all of the operations associated with the active
   * requests in the array have completed.
   * <p>
   * <table>
   * <tr><td><tt> array_of_requests </tt></td><td> array of requests </tr>
   * <tr><td><em> returns:          </em></td><td> array of status objects </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_WAITALL</tt>.
   * <p>
   * The result array will be the same size as <tt>array_of_requests</tt>.
   * On exit, requests become inactive.  If the <em>input</em> value of
   * <tt>arrayOfRequests</tt> contains inactive requests, corresponding
   * elements of the result array will contain null status handles.
   */

  private static native Status[] waitall(Request [] array_of_request);
  public static Status[] Waitall (Request [] array_of_request){ 
    Status result[] = waitall(array_of_request);
            
    int request_length = array_of_request.length;
     
    for (int i = 0; i<request_length;i++){  
      switch (array_of_request[i].opTag) {
        case Tag_Normal :
          break;
        case Tag_Isend_Object :
            array_of_request[i].dataReq.Wait(new Status());
          break;
        case Tag_Irecv_Object :
            byte[] byte_buf = new byte[array_of_request[i].length_buf[0]];
            result[i].object_count = array_of_request[i].length_buf[1]; 
            array_of_request[i].comm.Recv(byte_buf,0,
                   array_of_request[i].length_buf[0], MPI.BYTE,
                   array_of_request[i].src, array_of_request[i].tag);

            array_of_request[i].comm.Object_Deserialize(
                   array_of_request[i].buf,byte_buf,
                   array_of_request[i].offset,
                   array_of_request[i].length_buf[1],
                   array_of_request[i].type);
          break;
        case Tag_Send_Object :
            array_of_request[i].dataReq.Wait(new Status());
          break;
        case Tag_Bsend_Object :
            array_of_request[i].dataReq.Wait(new Status());
          break;
        case Tag_Ssend_Object :
            array_of_request[i].dataReq.Wait(new Status());
          break;
        case Tag_Rsend_Object :
            array_of_request[i].dataReq.Wait(new Status());
          break;
        case Tag_Recv_Object :
            array_of_request[i].comm.Recv(array_of_request[i].length_buf,
                                          0,2, MPI.INT,
                                          array_of_request[i].src, 
                                          array_of_request[i].tag);

            byte_buf = new byte[array_of_request[i].length_buf[0]]; 
            result[i].object_count = array_of_request[i].length_buf[1];
            array_of_request[i].comm.Recv(byte_buf,0,
                   array_of_request[i].length_buf[0], MPI.BYTE,
                   array_of_request[i].src, array_of_request[i].tag);

            array_of_request[i].comm.Object_Deserialize(
                   array_of_request[i].buf,byte_buf,
                   array_of_request[i].offset,
                   array_of_request[i].length_buf[1],
                   array_of_request[i].type);
          break;
      }
    }
    return result;
  }

  /**
   * Tests for completion of <em>all</em> of the operations associated
   * with active requests.
   * <p>
   * <table>
   * <tr><td><tt> array_of_requests </tt></td><td> array of requests </tr>
   * <tr><td><em> returns:          </em></td><td> array of status objects </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TESTALL</tt>.
   * <p>
   * If all operations have completed, the exit value of the argument array
   * and the result array are as for <tt>Waitall</tt>.  If any
   * operation has not completed, the result value is null and no
   * element of the argument array is modified.
   */

  private static native Status[] testall(Request [] array_of_request);
  public static Status[] Testall(Request [] array_of_request){
    Status result[] = testall(array_of_request);

    if (result == null)
      return null;

    int request_length = array_of_request.length;
    
    for (int i = 0; i<request_length;i++){ 
      switch (array_of_request[i].opTag) {
        case Tag_Normal :
          break;
        case Tag_Isend_Object :
            array_of_request[i].dataReq.Test(new Status());
          break;
        case Tag_Irecv_Object :
            byte[] byte_buf = new byte[array_of_request[i].length_buf[0]];

            result[i].object_count = array_of_request[i].length_buf[1]; 
            array_of_request[i].comm.Recv(byte_buf,0,
                   array_of_request[i].length_buf[0], MPI.BYTE,
                   array_of_request[i].src, array_of_request[i].tag);

            array_of_request[i].comm.Object_Deserialize(
                   array_of_request[i].buf,byte_buf,
                   array_of_request[i].offset,
                   array_of_request[i].length_buf[1],
                   array_of_request[i].type);
          break;
        case Tag_Send_Object :
            array_of_request[i].dataReq.Test(new Status());
          break;
        case Tag_Bsend_Object :
            array_of_request[i].dataReq.Test(new Status());
          break;
        case Tag_Ssend_Object :
            array_of_request[i].dataReq.Test(new Status());
          break;
        case Tag_Rsend_Object :
            array_of_request[i].dataReq.Test(new Status());
          break;
        case Tag_Recv_Object :
            array_of_request[i].comm.Recv(array_of_request[i].length_buf,
                                          0, 2, MPI.INT,
                                          array_of_request[i].src, 
                                          array_of_request[i].tag);

            byte_buf = new byte[array_of_request[i].length_buf[0]]; 
            result[i].object_count = array_of_request[i].length_buf[1];
            array_of_request[i].comm.Recv(byte_buf,0,
                   array_of_request[i].length_buf[0], MPI.BYTE,
                   array_of_request[i].src, array_of_request[i].tag);

            array_of_request[i].comm.Object_Deserialize(
                   array_of_request[i].buf,byte_buf,
                   array_of_request[i].offset,
                   array_of_request[i].length_buf[1],
                   array_of_request[i].type);
          break;
      }
    }
    return result;
  }

  /**
   * Blocks until at least one of the operations associated with the active
   * requests in the array has completed.
   * <p>
   * <table>
   * <tr><td><tt> array_of_requests </tt></td><td> array of requests </tr>
   * <tr><td><em> returns:          </em></td><td> array of status objects </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_WAITSOME</tt>.
   * <p>
   * The size of the result array will be the number of operations that
   * completed.  The index in <tt>array_of_requests</tt> for each request that
   * completed can be obtained from the returned status objects through the
   * <tt>Status.index</tt> field.  The corresponding element in
   * <tt>array_of_requests</tt> becomes inactive.
   */

  private static native Status[] waitsome(Request [] array_of_request);
  public static Status[] Waitsome(Request [] array_of_request){
    Status result[] = waitsome(array_of_request);                         
            
    int result_length = result.length;   
     
    for (int i = 0; i<result_length;i++){
      switch (array_of_request[result[i].index].opTag) {
        case Tag_Normal : 
          break;
        case Tag_Isend_Object :
            array_of_request[result[i].index].dataReq.Wait(new Status());
          break;
        case  Tag_Irecv_Object :
            byte[] byte_buf = new byte[array_of_request[result[i].index].length_buf[0]]; 
            result[i].object_count = array_of_request[result[i].index].length_buf[1];
            array_of_request[result[i].index].comm.Recv(byte_buf,0,
                   array_of_request[result[i].index].length_buf[0], MPI.BYTE,
                   array_of_request[result[i].index].src, 
                   array_of_request[result[i].index].tag);

            array_of_request[result[i].index].comm.Object_Deserialize(
                   array_of_request[result[i].index].buf,byte_buf,
                   array_of_request[result[i].index].offset,
                   array_of_request[result[i].index].length_buf[1],
                   array_of_request[result[i].index].type);
          break;
        case Tag_Send_Object :
            array_of_request[result[i].index].dataReq.Wait(new Status());
          break;
        case Tag_Bsend_Object :
            array_of_request[result[i].index].dataReq.Wait(new Status());
          break;
        case Tag_Ssend_Object :
            array_of_request[result[i].index].dataReq.Wait(new Status());
          break;
        case Tag_Rsend_Object :
            array_of_request[result[i].index].dataReq.Wait(new Status());
          break;
        case Tag_Recv_Object :
            array_of_request[result[i].index].comm.Recv(
                   array_of_request[result[i].index].length_buf,0,2,MPI.INT, 
                   array_of_request[result[i].index].src, 
                   array_of_request[result[i].index].tag);

            byte_buf = new byte[array_of_request[result[i].index].length_buf[0]]; 
            result[i].object_count = array_of_request[result[i].index].length_buf[1];
            array_of_request[result[i].index].comm.Recv(byte_buf,0,
                   array_of_request[result[i].index].length_buf[0], MPI.BYTE,
                   array_of_request[result[i].index].src,
                   array_of_request[result[i].index].tag);

            array_of_request[result[i].index].comm.Object_Deserialize(
                   array_of_request[result[i].index].buf,byte_buf,
                   array_of_request[result[i].index].offset,
                   array_of_request[result[i].index].length_buf[1],
                   array_of_request[result[i].index].type);
          break; 
      }
    }
    return result;    
  }

  /**
   * Behaves like <tt>Waitsome</tt>, except that it returns immediately.
   * <p>
   * <table>
   * <tr><td><tt> array_of_requests </tt></td><td> array of requests </tr>
   * <tr><td><em> returns:          </em></td><td> array of status objects </tr>
   * </table>
   * <p>
   * Java binding of the MPI operation <tt>MPI_TESTSOME</tt>.
   * <p>
   * If no operation has completed, <tt>TestSome</tt> returns an array of
   * length zero and elements of <tt>array_of_requests</tt> are unchanged.
   * Otherwise, arguments and return value are as for <tt>Waitsome</tt>.
   */

  private static native Status[] testsome(Request [] array_of_request);
  public static Status[] Testsome(Request [] array_of_request){
    Status result[] = testsome(array_of_request);

    if (result  == null)
      return null;

    int result_length = result.length;  
     
    for (int i = 0; i<result_length;i++){
      switch (array_of_request[result[i].index].opTag) {
        case Tag_Normal : 
          break;
        case Tag_Isend_Object :
            array_of_request[result[i].index].dataReq.Test(new Status());
            break;
        case Tag_Irecv_Object :
            byte[] byte_buf = new byte[array_of_request[result[i].index].length_buf[0]]; 
            result[i].object_count = array_of_request[result[i].index].length_buf[1];
            array_of_request[result[i].index].comm.Recv(byte_buf,0,
                   array_of_request[result[i].index].length_buf[0], MPI.BYTE,
                   array_of_request[result[i].index].src, 
                   array_of_request[result[i].index].tag);

            array_of_request[result[i].index].comm.Object_Deserialize(
                   array_of_request[result[i].index].buf,byte_buf,
                   array_of_request[result[i].index].offset,
                   array_of_request[result[i].index].length_buf[1],
                   array_of_request[result[i].index].type);
          break;
        case Tag_Send_Object :
            array_of_request[result[i].index].dataReq.Test(new Status());
          break;
        case Tag_Bsend_Object :
            array_of_request[result[i].index].dataReq.Test(new Status());
           break;
        case Tag_Ssend_Object :
            array_of_request[result[i].index].dataReq.Test(new Status());
           break;
        case Tag_Rsend_Object :
            array_of_request[result[i].index].dataReq.Test(new Status());
           break;
        case Tag_Recv_Object :
            array_of_request[result[i].index].comm.Recv(
                   array_of_request[result[i].index].length_buf,0,2,MPI.INT, 
                   array_of_request[result[i].index].src, 
                   array_of_request[result[i].index].tag);

            byte_buf = new byte[array_of_request[result[i].index].length_buf[0]]; 
            result[i].object_count = array_of_request[result[i].index].length_buf[1];
            array_of_request[result[i].index].comm.Recv(byte_buf,0,
                   array_of_request[result[i].index].length_buf[0], MPI.BYTE,
                   array_of_request[result[i].index].src,
                   array_of_request[result[i].index].tag);

            array_of_request[result[i].index].comm.Object_Deserialize(
                   array_of_request[result[i].index].buf,byte_buf,
                   array_of_request[result[i].index].offset,
                   array_of_request[result[i].index].length_buf[1],
                   array_of_request[result[i].index].type);
          break; 
      }
    }
    return result;    
  }
  protected long handle;

  static {
    init();
  }          
}
